"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.UserEvent = exports.User = void 0;
var _typedEventEmitter = require("./typed-event-emitter");
function _defineProperty(obj, key, value) { key = _toPropertyKey(key); if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }
function _toPropertyKey(arg) { var key = _toPrimitive(arg, "string"); return typeof key === "symbol" ? key : String(key); }
function _toPrimitive(input, hint) { if (typeof input !== "object" || input === null) return input; var prim = input[Symbol.toPrimitive]; if (prim !== undefined) { var res = prim.call(input, hint || "default"); if (typeof res !== "object") return res; throw new TypeError("@@toPrimitive must return a primitive value."); } return (hint === "string" ? String : Number)(input); }
let UserEvent;
exports.UserEvent = UserEvent;
(function (UserEvent) {
  UserEvent["DisplayName"] = "User.displayName";
  UserEvent["AvatarUrl"] = "User.avatarUrl";
  UserEvent["Presence"] = "User.presence";
  UserEvent["CurrentlyActive"] = "User.currentlyActive";
  UserEvent["LastPresenceTs"] = "User.lastPresenceTs";
})(UserEvent || (exports.UserEvent = UserEvent = {}));
class User extends _typedEventEmitter.TypedEventEmitter {
  /**
   * The 'displayname' of the user if known.
   * @privateRemarks
   * Should be read-only
   */

  /**
   * The 'avatar_url' of the user if known.
   * @privateRemarks
   * Should be read-only
   */

  /**
   * The presence status message if known.
   * @privateRemarks
   * Should be read-only
   */

  /**
   * The presence enum if known.
   * @privateRemarks
   * Should be read-only
   */

  /**
   * Timestamp (ms since the epoch) for when we last received presence data for this user.
   * We can subtract lastActiveAgo from this to approximate an absolute value for when a user was last active.
   * @privateRemarks
   * Should be read-only
   */

  /**
   * The time elapsed in ms since the user interacted proactively with the server,
   * or we saw a message from the user
   * @privateRemarks
   * Should be read-only
   */

  /**
   * Whether we should consider lastActiveAgo to be an approximation
   * and that the user should be seen as active 'now'
   * @privateRemarks
   * Should be read-only
   */

  /**
   * The events describing this user.
   * @privateRemarks
   * Should be read-only
   */

  /**
   * Construct a new User. A User must have an ID and can optionally have extra information associated with it.
   * @param userId - Required. The ID of this user.
   */
  constructor(userId) {
    super();
    this.userId = userId;
    _defineProperty(this, "modified", -1);
    _defineProperty(this, "displayName", void 0);
    _defineProperty(this, "rawDisplayName", void 0);
    _defineProperty(this, "avatarUrl", void 0);
    _defineProperty(this, "presenceStatusMsg", void 0);
    _defineProperty(this, "presence", "offline");
    _defineProperty(this, "lastActiveAgo", 0);
    _defineProperty(this, "lastPresenceTs", 0);
    _defineProperty(this, "currentlyActive", false);
    _defineProperty(this, "events", {});
    this.displayName = userId;
    this.rawDisplayName = userId;
    this.updateModifiedTime();
  }

  /**
   * Update this User with the given presence event. May fire "User.presence",
   * "User.avatarUrl" and/or "User.displayName" if this event updates this user's
   * properties.
   * @param event - The `m.presence` event.
   *
   * @remarks
   * Fires {@link UserEvent.Presence}
   * Fires {@link UserEvent.DisplayName}
   * Fires {@link UserEvent.AvatarUrl}
   */
  setPresenceEvent(event) {
    if (event.getType() !== "m.presence") {
      return;
    }
    const firstFire = this.events.presence === null;
    this.events.presence = event;
    const eventsToFire = [];
    if (event.getContent().presence !== this.presence || firstFire) {
      eventsToFire.push(UserEvent.Presence);
    }
    if (event.getContent().avatar_url && event.getContent().avatar_url !== this.avatarUrl) {
      eventsToFire.push(UserEvent.AvatarUrl);
    }
    if (event.getContent().displayname && event.getContent().displayname !== this.displayName) {
      eventsToFire.push(UserEvent.DisplayName);
    }
    if (event.getContent().currently_active !== undefined && event.getContent().currently_active !== this.currentlyActive) {
      eventsToFire.push(UserEvent.CurrentlyActive);
    }
    this.presence = event.getContent().presence;
    eventsToFire.push(UserEvent.LastPresenceTs);
    if (event.getContent().status_msg) {
      this.presenceStatusMsg = event.getContent().status_msg;
    }
    if (event.getContent().displayname) {
      this.displayName = event.getContent().displayname;
    }
    if (event.getContent().avatar_url) {
      this.avatarUrl = event.getContent().avatar_url;
    }
    this.lastActiveAgo = event.getContent().last_active_ago;
    this.lastPresenceTs = Date.now();
    this.currentlyActive = event.getContent().currently_active;
    this.updateModifiedTime();
    for (const eventToFire of eventsToFire) {
      this.emit(eventToFire, event, this);
    }
  }

  /**
   * Manually set this user's display name. No event is emitted in response to this
   * as there is no underlying MatrixEvent to emit with.
   * @param name - The new display name.
   */
  setDisplayName(name) {
    const oldName = this.displayName;
    this.displayName = name;
    if (name !== oldName) {
      this.updateModifiedTime();
    }
  }

  /**
   * Manually set this user's non-disambiguated display name. No event is emitted
   * in response to this as there is no underlying MatrixEvent to emit with.
   * @param name - The new display name.
   */
  setRawDisplayName(name) {
    this.rawDisplayName = name;
  }

  /**
   * Manually set this user's avatar URL. No event is emitted in response to this
   * as there is no underlying MatrixEvent to emit with.
   * @param url - The new avatar URL.
   */
  setAvatarUrl(url) {
    const oldUrl = this.avatarUrl;
    this.avatarUrl = url;
    if (url !== oldUrl) {
      this.updateModifiedTime();
    }
  }

  /**
   * Update the last modified time to the current time.
   */
  updateModifiedTime() {
    this.modified = Date.now();
  }

  /**
   * Get the timestamp when this User was last updated. This timestamp is
   * updated when this User receives a new Presence event which has updated a
   * property on this object. It is updated <i>before</i> firing events.
   * @returns The timestamp
   */
  getLastModifiedTime() {
    return this.modified;
  }

  /**
   * Get the absolute timestamp when this User was last known active on the server.
   * It is *NOT* accurate if this.currentlyActive is true.
   * @returns The timestamp
   */
  getLastActiveTs() {
    return this.lastPresenceTs - this.lastActiveAgo;
  }
}
exports.User = User;